/* zgsmplay 1.4 - a real-time curses interface for playing gsm files.
 * Copyright (C) 1995-2004 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * zgsmplay.c - main program and the actual gsm player.
 */

#define ZGSMPLAY_VER	"1.4"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <curses.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <termios.h>		/* for struct winsize and TIOCGWINSZ */
#include <gsm.h>

#include "filesel.h"
#include "readnbkey.h"
#include "zgsmplay.h"


/* switch to seconds-only mode for files bigger than 15 mins */
#define SEC_LIMIT	(15*60)

/* ,..and to minutes-only mode for files bigger than 10 hours */
#define MINUTES_LIMIT	(10*60*60)

/* the N used when playing every Nth frame during audio-search */
#define FAST_PLAY_EVERY_NTH	6


/* 1024 byte buffer seems to work well enough and not tax the cpu
 * too heavily. this is for 8kHz; it's doubled for each doubling of
 * the sample rate above 8kHz.
 *
 * (Or rather, it used to be - variable-sample-rate isn't supported any
 * more. It *is* still doubled when using 16-bit or stereo, however.)
 */
#define BASE_FRAG_POWER		10
#define BASE_ACCUM_BUFSIZ	(1<<BASE_FRAG_POWER)
#define MAX_ACCUM_BUFSIZ	16384

#define AUDIO_DEV	"/dev/dsp"


int frag_power=BASE_FRAG_POWER;
int accum_bufsiz=BASE_ACCUM_BUFSIZ;
int start_frame=0;
int usefrag=1;
int need_stereo=0,pstereo=0;
int pstereobuf[1024];
int pstereobufsiz,pstereopos;
int psgap=250;

int ttyfd;

int sgres,pic_incr;


/* sample rate (frequency) for GSM playback.
 *
 * NB: This is *NOT* used everywhere - to avoid problems with large files
 * (as in, more than about 65 megs), some time calculations use e.g.
 * filepos/1650 instead of filepos*160/33/8000. Also, since /dev/dsp's
 * playback rate is not explicitly set, you'd need to add an ioctl() to
 * do that too before changing this would really work.
 */
int freq=8000;

/* if true, using 16-bit playback. */
int sixteenbit=1;


char *filename=NULL;
gsm gsmhndl;
gsm_signal outbuf[160];		/* gsm_signal = signed short */
unsigned char accumbuf[MAX_ACCUM_BUFSIZ*4]; /* x4 in case of 16-bit stereo */
int lastaccum;
gsm_frame inbuf;

int autorepeat=0;
int bar_xpos=3;
int bar_ypos=4;
int line_chars=1;	/* set if line graphics chars should be used */

/* XXX should have line_chars as cmd-line option, as ncurses's replacements
 * for the ACS_* chars kind of suck; they may be logical and general,
 * but they just don't look as nice as the old zgsmplay flat-ascii boxology.
 */


volatile int window_resized=0;	/* set if SIGWINCH received */
int term_too_small=0;

/* complaint displayed in top-left if window is resized to be too small
 * for zgsmplay to work. This should be displayed each time round the
 * loop if term_too_small is true, to maximise the chance it won't
 * get scrolled off etc.
 *
 * It's phrased oddly so that the "78x7" is likely to be visible no matter
 * what the window size.
 */
char *term_small_complaint="*** need >=78x7 - window too small! ***";

#ifdef SET_WINDOW_NAME
int have_xterm=0;
#endif

/* the char to use for the bar when line_chars is true;
 * I experimented with ACS_CKBOARD, but I just didn't like the look of
 * it somehow. Define USE_GREY_CHAR to enable it anyway.
 */
#ifdef USE_GREY_CHAR
#define GREY_CHAR	ACS_CKBOARD
#else
#define GREY_CHAR	'#'
#endif



void clearscreen(void)
{
wclear(stdscr);
move(0,0); refresh();
}


void resize_handler(int a)
{
window_resized=1;
}


void screenon(void)
{
struct sigaction sa;

initscr();
cbreak(); noecho();
curs_set(0);	/* remove cursor (if possible) */

/* catch resize signals.
 * It would be easier to use ncurses's handy KEY_RESIZE for this,
 * but that would mean changing readnbkey(), and would generally be
 * a pain in the bum. :-/
 */
sigemptyset(&sa.sa_mask);
sa.sa_handler=resize_handler;
sa.sa_flags=SA_RESTART;
sigaction(SIGWINCH,&sa,NULL);

if((ttyfd=open("/dev/tty",O_RDONLY|O_NONBLOCK))<0)
  {
  screenoff();
  fprintf(stderr,"zgsmplay: couldn't open /dev/tty!\n");
  exit(1);
  }

/* check term size */
if(COLS<78 || LINES<7)
  {
  screenoff();
  fprintf(stderr,"zgsmplay: "
  		"sorry, need a terminal with at least 78 cols and 7 lines.\n");
  exit(1);
  }

#ifdef SET_WINDOW_NAME
have_xterm=0;
if(getenv("TERM") &&
   (strncmp(getenv("TERM"),"xterm",5)==0 ||
    strncmp(getenv("TERM"),"rxvt",4)==0))
  {
  have_xterm=1;
  fflush(stdout);	/* just in case */
  /* set window name to `zgsmplay' */
  fprintf(stderr,"\x1b]0;zgsmplay\x07");
  }
#endif
}


void screenoff(void)
{
#ifdef SET_WINDOW_NAME
char *ptr;
#endif

if(ttyfd>=0) close(ttyfd);
clearscreen();
move(LINES-1,0);
refresh();
echo(); nocbreak();
endwin();
putchar('\n');

/* in case of error exit, since stderr will be used immediately after... */
fflush(stdout);

#ifdef SET_WINDOW_NAME
if(have_xterm)
  {
  /* reset window name to `xterm'/`rxvt' */
  ptr="xterm";
  /* see rxvt FAQ for rationale... */
  if(getenv("COLORTERM") && strcmp(getenv("COLORTERM"),"rxvt")==0)
    ptr="rxvt";
  
  fprintf(stderr,"\x1b]0;%s\x07",ptr);
  }
#endif
}


/* put cursor out of the way - call just before refresh().
 * If term has been resized to be too small, complains about that instead.
 */
void stashcursor(void)
{
if(term_too_small)
  mvaddstr(0,0,term_small_complaint);
else
  move(LINES-1,0);
}


void usage_help(void)
{
printf("zgsmplay %s - copyright (c) 1995-2004 Russell Marks.\n\n",
	ZGSMPLAY_VER);
printf("usage: zgsmplay [-hnrS] [-f frame] [-g gap]\n"
"			[start_dir | filename.gsm | filename1.gsm ...]\n"
"\n"
"	-f	set GSM frame to start from (default is 0), or (with `s'\n"
"		or `m' suffix) second/minute to start from.\n"
"	-g	set pseudo-stereo echo length (`gap'). The units are\n"
"		number of samples at 22kHz and are scaled accordingly.\n"
"		The default is 250, and the maximum acceptable is 1024.\n"
"	-h	this usage help.\n"
"	-n	no snd_frag. Takes much longer to respond to keypresses\n"
"		(several seconds!) but breaks up *much* less, usually not\n"
"		at all, on a heavily-loaded machine.\n"
"	-r	enable auto-repeat.\n"
"	-S	use pseudo-stereo, which often sounds good with music.\n"
"		(It can make speech sound rather odd, though.)\n"
"\n"
"  start_dir	directory to start file selector in.\n"
"  filename.gsm	single file to play; skips file selector entirely.\n"
"  filename1.gsm ...\n"
"		a number of files to play in order specified; also skips\n"
"		file selector.\n"
"\n"
"If standard input is not a terminal, zgsmplay non-interactively plays\n"
"GSM data directly from stdin. Such usage implies `-n'.\n");
}


/* parse cmdline and setup any specified file for reading.
 * (note that we can't do any curses stuff here; nor should we want to ;-))
 * returns number of args left after parsing opts.
 */
int init(int argc,char *argv[])
{
int done=0,tmp;

/* default to no filename */
filename=NULL;

opterr=0;

do
  switch(getopt(argc,argv,"f:g:hnrsS"))
    {
    case 'f':	/* start frame */
      start_frame=atoi(optarg);
      /* XXX this is pretty crude :-) */
      if(strlen(optarg)>=2)
        switch(optarg[strlen(optarg)-1])
          {
          case 'm':
            start_frame*=60;
            /* FALLS THROUGH */
          case 's':
            start_frame*=50;	/* 50 frames/sec in GSM */
          }
      if(start_frame<0) start_frame=0;
      break;
    case 'g':	/* ps gap */
      tmp=atoi(optarg);
      if(tmp==0 || tmp>1024)
        fprintf(stderr,"zgsmplay: strange gap length, ignoring\n");
      else
        psgap=tmp;
      break;
    case 'n':	/* no snd_frag */
      usefrag=0;
      break;
    case 'h':
      usage_help();
      exit(1);
    case 'r':	/* enable auto-repeat */
      autorepeat=1;
      break;
    case 'S':	/* pseudo-stereo */
    case 's':	/* tolerate this as alternative */
      pstereo=1;
      sixteenbit=0;	/* XXX only allowed in 8-bit for now */
      break;
    case '?':
      switch(optopt)
        {
        case 'f':
          fprintf(stderr,"zgsmplay: "
          	"the -f option needs a frame number as argument.\n");
          break;
        case 'g':
          fprintf(stderr,"zgsmplay: the -g option needs a gap as argument.\n");
          break;
        default:
          fprintf(stderr,"zgsmplay: option `%c' not recognised.\n",optopt);
        }
      exit(1);
    case -1:
      done=1;
    }
while(!done);

return(argc-optind);	/* num of args not accounted for */
}


/* curses has the `border'-drawing routines, but they work on windows
 * only, and I tend not to make much use of windows. So here's a simple
 * box-drawing routine. It assumes width and height are both >=3.
 */
void draw_box(int y,int x,int w,int h,int horiz,int vert,
	int ul,int ur,int ll,int lr)
{
mvaddch(y,x,ul);	/* ul corner */
hline(horiz,w-2);	/* top line */
mvaddch(y,x+w-1,ur);	/* ur corner */

move(y+1,x);
vline(vert,h-2);	/* left line */
move(y+1,x+w-1);
vline(vert,h-2);	/* right line */

mvaddch(y+h-1,x,ll);	/* ll corner */
hline(horiz,w-2);	/* bottom line */
mvaddch(y+h-1,x+w-1,lr); /* lr corner */
}


/* draw header/footer stuff, which doesn't change while playing.
 * Also initialises position of bar (bar_[xy]pos), which most positions
 * are calculated relative to.
 */
void drawheadfoot(int filesize)
{
/* this is ((filesize*160)/33)/freq, hardcoded for 8kHz so we can lose
 * the multiplication and thus cope with *big* files (up to 2 gigs,
 * in practice).
 */
unsigned int secsize=filesize/1650;	/* expanded len in sec */
unsigned int f,xpos,tmp;
char *name;

if(LINES>20)
  bar_ypos=LINES/2-2;	/* allow for help display */
else
  if(LINES>11)
    bar_ypos=LINES/2-1;
  else
    bar_ypos=4;		/* works for 10 or 11 lines */

bar_xpos=3;
if(COLS>77)		/* yes, 77 :-) */
  bar_xpos+=(COLS-77)/2;

/* show filename centred at top */
name=filename;
if(strrchr(filename,'/'))
  name=strrchr(filename,'/')+1;
tmp=strlen(name);
if(tmp>COLS) tmp=COLS;
mvaddnstr(0,(COLS-tmp)/2,name,COLS);

if(line_chars)
  draw_box(bar_ypos,bar_xpos,71,3,
    ACS_HLINE,ACS_VLINE,
    ACS_ULCORNER,ACS_URCORNER,
    ACS_LLCORNER,ACS_LRCORNER);
else
  draw_box(bar_ypos,bar_xpos,71,3, '-','|', ',','.', '`','\'');

/* only do the tick marks/times if we have space to fit both in */
if(bar_ypos+4<=LINES-1)
  {
  /* ACS_VLINE doesn't look right for this, so it's always `|'. */
  for(f=0;f<=71;f+=10)
    mvaddch(bar_ypos+3,bar_xpos+f,'|');
  
  /* draw times at tick marks above */
  for(f=0,xpos=0;f<8;f++,xpos+=10)
    {
    /* time mark of this position */
    tmp=(f*secsize)/7;
    move(bar_ypos+4,bar_xpos-3+xpos);
    if(secsize<MINUTES_LIMIT)
      printw("%1d:%02d:%02d",tmp/3600,(tmp%3600)/60,tmp%60);
    else
      printw("%3d:%02d",tmp/3600,(tmp%3600)/60);
    }
  }

if(LINES>20)
  {
  mvaddstr(LINES-3,bar_xpos+1,
    "Esc - exit              p - play/pause          R - toggle auto-repeat");
  mvaddstr(LINES-2,bar_xpos+1,
    "h - back 15 sec         j - back 1%             r - rewind");
  mvaddstr(LINES-1,bar_xpos+1,
    "l - forward 15 sec      k - forward 1%          e - go to end");
  }
}


/* draw place in file */
void drawplace(FILE *in,int filesize,int playing)
{
int curpos=ftell(in);
/* see note at top of drawheadfoot() for explanation of next 2 lines... */
unsigned int secsize=filesize/1650;	/* expanded len in sec */
unsigned int secpos=curpos/1650;	/* similar to current pos */
unsigned int f,tmp;
int hashchar=line_chars?GREY_CHAR:'#';

move(bar_ypos-2,bar_xpos-3);

if(secsize<SEC_LIMIT)	/* if under 15 mins long, use 1/100-th sec */
  {
  /* the *100 is no problem, it's known to be a `small' file */
  unsigned int hsecsize=(filesize*100)/1650;	/* in 1/100 sec */
  unsigned int hsecpos=(curpos*100)/1650;
  
  printw("%3d%%, %1d:%02d:%02d.%02d of %1d:%02d:%02d.%02d",
    (hsecpos*100)/hsecsize,
    hsecpos /360000,(hsecpos %360000)/6000,(hsecpos %6000)/100,hsecpos %100,
    hsecsize/360000,(hsecsize%360000)/6000,(hsecsize%6000)/100,hsecsize%100);
  
  tmp=(69*hsecpos)/hsecsize;
  }
else
  {
  if(secsize<MINUTES_LIMIT)
    {
    /* use seconds res. only */
    printw("%3d%%, %1d:%02d:%02d of %1d:%02d:%02d",
      (secpos*100)/secsize,
      secpos /3600,(secpos %3600)/60,secpos %60,
      secsize/3600,(secsize%3600)/60,secsize%60);
    }
  else
    {
    /* use minutes res. only */
    printw("%3d%%, %3d:%02d of %3d:%02d",
      (secpos*100)/secsize,
      secpos /3600,(secpos %3600)/60,
      secsize/3600,(secsize%3600)/60);
    }
  
  tmp=(69*secpos)/secsize;
  }

printw(", frame %08d",curpos/33);

/* just in case... */
if(tmp<0) tmp=0;
if(tmp>69) tmp=69;
  
move(bar_ypos+1,bar_xpos+1);

for(f=0;f<69;f++)
#ifndef OLD_BAR_STYLE
  addch((f<tmp)?hashchar:' ');	/* accurate */
#else
  addch((f<=tmp)?hashchar:' ');	/* alternative (old) approach */
#endif

mvaddstr(bar_ypos-2,bar_xpos+62,playing?"[ PLAYING ]":"[ STOPPED ]");
mvaddstr(bar_ypos-2,bar_xpos+51,autorepeat?"(repeat)":"        ");
}


void output_pstereo(int c,unsigned char *outl,unsigned char *outr)
{
int bl=128+(c-128-pstereobuf[pstereopos])/2;
int br=128+(c-128+pstereobuf[pstereopos])/2;

if(bl<0) bl=0;
if(br<0) br=0;
if(bl>255) bl=255;
if(br>255) br=255;

*outl=(unsigned char)bl; *outr=(unsigned char)br;
  
pstereobuf[pstereopos]=c;
pstereopos++; if(pstereopos>=pstereobufsiz) pstereopos=0;
}


/* play N gsm frames
 * returns non-zero if ok, zero on eof
 */
int playframes(FILE *in,int out,int n)
{
int f,g,i;

for(f=0;f<n;f++)
  {
  if(fread(inbuf,1,sizeof(inbuf),in)==sizeof(inbuf))
    if(gsm_decode(gsmhndl,inbuf,outbuf)>=0)
      {
      /* converts signed words to unsigned bytes if needed */
      
      for(g=0;g<160;g++)
        {
        if(sixteenbit)
          {
          /* assumes little-endian, but we *did* explicitly ask for that :-) */
          accumbuf[lastaccum++]=outbuf[g]&255;
          accumbuf[lastaccum++]=(outbuf[g]>>8);
          if(need_stereo)
            {
            accumbuf[lastaccum++]=outbuf[g]&255;
            accumbuf[lastaccum++]=(outbuf[g]>>8);
            }
          }
        else
          {
          /* 8-bit */
          i=(int)outbuf[g]+32768;
          i>>=8;
          if(pstereo)
            {
            output_pstereo(i,accumbuf+lastaccum,accumbuf+lastaccum+1);
            lastaccum+=2;
            }
          else
            {
            accumbuf[lastaccum++]=(unsigned char)i;
            if(need_stereo)
              accumbuf[lastaccum++]=(unsigned char)i;
            }
          }
        
        /* if accumbuf[] is full, write it */
        if(lastaccum>=accum_bufsiz)
          {
          write(out,accumbuf,accum_bufsiz);
          lastaccum=0;
          }
        }
      }
  }

if(feof(in)) return(0);
return(1);
}


long getfilesize(FILE *in)
{
struct stat sbuf;

if(fstat(fileno(in),&sbuf)==-1)
  return(-1);

return((long)sbuf.st_size);
}


void relseek(FILE *in,int filesize,int add)
{
int delta=(add>0)?add:-add;
int curpos=ftell(in);

/* check we're not seeking back past 0 */
if(add<0 && curpos<delta) add=-curpos;

/* check for other end */
if(add>0 && filesize-curpos<delta) add=filesize-curpos;

fseek(in,curpos+add,SEEK_SET);
}


/* fix LINES and COLS (and ncurses) based on current terminal size.
 * Called after a SIGWINCH.
 */
void fix_scrn_size(void)
{
struct winsize ws;

ioctl(0,TIOCGWINSZ,&ws);
LINES=ws.ws_row;
COLS=ws.ws_col;

/* let ncurses know about it */
resizeterm(LINES,COLS);

/* it's probably not a good idea to die if the terminal gets too small,
 * so pretend it's at least minimum size and struggle on. :-) But also
 * set flag so we get a `terminal too small' message.
 */
term_too_small=0;
if(COLS<78) COLS=78,term_too_small=1;
if(LINES<7) LINES=7,term_too_small=1;

/* turn off cursor again, as resizing in rxvt re-enables it;
 * however, since ncurses thinks it's still off, we have to enable it
 * first. This sucks a bit, but it's necessary. :-(
 */
curs_set(1);
curs_set(0);
}


void setup_stereo(int audiofd)
{
int f;

need_stereo=pstereo;
ioctl(audiofd,SNDCTL_DSP_STEREO,&need_stereo);
if(pstereo)
  {
  for(f=0;f<1024;f++)
    pstereobuf[f]=0;
  pstereopos=0;
  pstereobufsiz=(freq*psgap)/22000;
  }
}


/* play/navigate gsm file
 * returns non-zero if ok, zero on error
 */
int gsmplay(void)
{
FILE *in;
int out;
int quit=0;
int fifthsec=freq/160/5;	/* num of gsm frames which last 1/5 sec */
int sec15,onepcnt;		/* size of 15 sec and 1% in gsm frames */
int filesize;
int frag;
int playing=1;
int key;
int fastplay=0,fastcount=0;	/* for audio-search (HJKL) */
int rate=8000;

sgres=pic_incr=0;

/* open file, open audio, setup audio, create gsm handle */
if((in=fopen(filename,"rb"))==NULL)
  return(0);

/* if zero-length, return successfully.
 * (Important since the position-in-file messes up with zero len :-))
 */
if((filesize=getfilesize(in))==0)
  {
  fclose(in);
  return(1);
  }

if((out=open(AUDIO_DEV,O_WRONLY))<0)
  return(0);

frag_power=BASE_FRAG_POWER;
accum_bufsiz=BASE_ACCUM_BUFSIZ;
/* double assuming we'll get 16-bit (best we can do :-/) */
frag_power++;
accum_bufsiz*=2;

/* use 16-bit if possible and it hasn't previously failed */
if(sixteenbit)
  {
  int tmp=AFMT_S16_LE;
  
  if((ioctl(out,SNDCTL_DSP_SETFMT,&tmp))==-1)
    sixteenbit=0;
  }

/* use stereo if required */
setup_stereo(out);
if(need_stereo)
  {
  /* will want larger frag */
  frag_power++;
  accum_bufsiz*=2;
  }

/* set speed - which *shouldn't* be required for 8kHz, but it
 * seems some people have forgotten this. :-( Ignores errors,
 * as there's not much we could do about them anyhow.
 */
ioctl(out,SNDCTL_DSP_SPEED,&rate);

/* setup frag */
if(usefrag)
  {
#ifdef ASSUME_SLOW_MEDIA
  frag=0x80000|frag_power;	/* makes it pretty unresponsive... */
#else
  frag=0x20000|frag_power;
#endif
  ioctl(out,SNDCTL_DSP_SETFRAGMENT,&frag);
  }

if(!(gsmhndl=gsm_create())) return(0);

lastaccum=0;

if(filesize==0)
  {
  
  gsm_destroy(gsmhndl);
  close(out);
  fclose(in);
  return(1);
  }

sec15=(15*freq)/160;
onepcnt=(filesize/100)/33;
if(onepcnt==0) onepcnt++;	/* always make "1%" at least one frame */

clearscreen();
drawheadfoot(filesize);
drawplace(in,filesize,playing);
stashcursor();
refresh();

/* move to start position */
fseek(in,start_frame*33,SEEK_SET);

while(!quit && sgres==0)
  {
  if(!playing)
    {
    /* not playing, insert a delay to avoid busy looping */
    usleep(200000);
    }
  else
    {
    /* it's playing, so use the playback itself for an (approx) 1/5 sec
     * delay.
     */
    if(playframes(in,out,fastplay?1:fifthsec))
      {
      /* do fastplay frame skips if needed */
      if(fastplay)
        {
        fastcount-=FAST_PLAY_EVERY_NTH;
        if(fastcount<=0)
          fastplay=fastcount=0;
        else
          if(fastplay>0)
            /* playing it did a +1, add the rest */
            relseek(in,filesize,33*(FAST_PLAY_EVERY_NTH-1));
          else
            /* playing it did a +1, subtract one more to compensate,
             * and a further one so we're ready to play the right frame */
            relseek(in,filesize,-33*(FAST_PLAY_EVERY_NTH+2));
        }
      }
    else	/* if playframes() failed... */
      {
      /* if it reaches the end, we're done unless auto-repeating, in which
       * case just rewind (i.e. leave it playing).
       */
      if(autorepeat && !tagview_mode)
        rewind(in);
      else
        {
        /* flush accumbuf */
        if(lastaccum)
          {
          write(out,accumbuf,lastaccum);
          lastaccum=0;
          }
        
        playing=0;
        quit=1;
        }
      }
    }
  
  /* if we received SIGWINCH, redraw `frame' before updating stuff.
   * There's a race condition here of course, but I doubt it's
   * likely to be a problem in practice.
   */
  if(window_resized)
    {
    window_resized=0;
    fix_scrn_size();
    clearscreen();
    drawheadfoot(filesize);
    }
  
  /* update place-in-file bits */
  drawplace(in,filesize,playing);
  stashcursor();
  refresh();
  
  /* now check for any keys pressed, etc. */
  key=readnbkey();
  
  /* all keys other than quit and pause are ignored while fastplay
   * is in effect.
   */
  if(fastplay && key!=RK_ESC && key!='x' && key!='p' && key!='s')
    key=RK_NO_KEY;
  
  switch(key)
    {
    case 'h': case RK_CURSOR_LEFT:
      relseek(in,filesize,-sec15*33); break;
    case 'H':
      fastplay=-1; fastcount=sec15; break;
    case 'l': case RK_CURSOR_RIGHT:
      relseek(in,filesize, sec15*33); break;
    case 'L':
      fastplay=1; fastcount=sec15; break;
    case 'j': case RK_CURSOR_DOWN:
      relseek(in,filesize,-onepcnt*33); break;
    case 'J':
      fastplay=-1; fastcount=onepcnt; break;
    case 'k': case RK_CURSOR_UP:
      relseek(in,filesize, onepcnt*33); break;
    case 'K':
      fastplay=1; fastcount=onepcnt; break;
    case 'r': case RK_HOME:
      rewind(in); break;
    case 'e': case RK_END:
      fseek(in,0,SEEK_END); playing=0; lastaccum=0; break;
    case 'R': autorepeat=!autorepeat; break;
    case 's':
      playing=0; lastaccum=0; fastplay=fastcount=0; break;
    case 'p':
      playing=!playing; lastaccum=0; fastplay=fastcount=0; break;
    case 'N'-0x40:
      pic_incr=1;  sgres=1; break;
    case 'P'-0x40:
      pic_incr=-1; sgres=1; break;
    case RK_ENTER: case 13:
      pic_incr=1;  sgres=2; break;
    case 127: case 8:
      pic_incr=-1; sgres=2; break;
    case ' ':
      pic_incr=1;  sgres=3; break;
    case '1': case '2': case '3':
    case '4': case '5': case '6':
    case '7': case '8': case '9':
      /* move to n/10 of the way through the file;
       * the /33 then *33 is needed so that we do the maths in terms
       * of GSM frames.
       */
      fseek(in,(((filesize/33)*(key-'0'))/10)*33,SEEK_SET);
      break;
    
    case RK_ESC: case 'x':
      if(tagview_mode) tagview_mode=0;
      quit=1; sgres=1;
      break;
    }
  }

gsm_destroy(gsmhndl);
close(out);
fclose(in);
return(1);
}


/* play gsm file from stdin, like `toast -d' but with output to /dev/dsp.
 * This works non-interactively, and just saves having to otherwise use both
 * `sox' and `toast' for this purpose.
 * 
 * It's a bit messy having a separate function for this, but in practice
 * it works out easier this way.
 */
int gsmplay_toastlike(void)
{
int out;
int fifthsec=freq/160/5;	/* num of gsm frames which last 1/5 sec */

/* open audio, setup audio, create gsm handle */
if((out=open(AUDIO_DEV,O_WRONLY))<0)
  return(0);

/* use stereo if required */
setup_stereo(out);

/* use 16-bit if possible */
if(sixteenbit)
  {
  int tmp=AFMT_S16_LE;
  
  if((ioctl(out,SNDCTL_DSP_SETFMT,&tmp))==-1)
    sixteenbit=0;
  }

/* frag never used for playing stdin (no reason to, it's non-interactive) */

if(!(gsmhndl=gsm_create())) return(0);

lastaccum=0;

/* play (most of) the GSM */
while(playframes(stdin,out,fifthsec)) ;

/* flush accumbuf (i.e. play the rest :-)) */
if(lastaccum)
  write(out,accumbuf,lastaccum);

gsm_destroy(gsmhndl);
close(out);
return(1);
}



int main(int argc,char *argv[])
{
int argsleft=init(argc,argv);

/* if stdin isn't a tty, the GSM data to play comes from there. */
if(!isatty(0))
  {
  if(!gsmplay_toastlike())
    fprintf(stderr,"zgsmplay: /dev/dsp not writable\n"),exit(1);
  exit(0);
  }

gsmdir_init();		/* allocate initial memory for filesel's gsmdir[] */

if(argsleft==1)
  {
  struct stat sbuf;
  
  filename=argv[argc-1];
  
  /* check if it's a dir */
  if(stat(filename,&sbuf)!=-1 && S_ISDIR(sbuf.st_mode))
    chdir(filename);	/* ok, it is, just change dir and carry on */
  else
    {
    int res;
    /* one file specified; play it and quit */
    screenon();
    res=gsmplay();
    screenoff();
    if(res==0)
      fprintf(stderr,"zgsmplay: file not found (or /dev/dsp not writable)\n"),
      exit(1);
    exit(0);
    }
  }

if(argsleft>1)
  {
  /* multiple files specified; play them as a `slideshow' */
  /* see filesel.c for play_multi() */
  screenon();
  play_multi(argc,argv,argsleft);
  screenoff();
  exit(0);
  }

/* otherwise, normal startup */

screenon();
filesel();
screenoff();

exit(0);
}
